@@ -2,33 +2,39 @@
 # encoding: utf-8
 
 __author__ = "ChenyangGao <https://chenyanggao.github.io>"
-__version__ = (0, 1, 6)
+__version__ = (0, 1, 10)
 __all__ = [
     "Return", "Yield", "YieldFrom", "iterable", "async_iterable", "foreach", "async_foreach", 
-    "through", "async_through", "flatten", "async_flatten", "map", "filter", "reduce", "zip", 
-    "chunked", "iter_unique", "async_iter_unique", "wrap_iter", "wrap_aiter", "acc_step", 
-    "cut_iter", "run_gen_step", "run_gen_step_iter", "as_gen_step", "bfs_gen", "with_iter_next", 
-    "backgroud_loop", 
+    "through", "async_through", "flatten", "async_flatten", "collect", "async_collect", 
+    "group_collect", "async_group_collect", "map", "filter", "reduce", "zip", "chunked", 
+    "iter_unique", "async_iter_unique", "wrap_iter", "wrap_aiter", "acc_step", "cut_iter", 
+    "iter_gen_step", "iter_gen_step_async", "run_gen_step", "run_gen_step_iter", 
+    "as_gen_step", "bfs_gen", "with_iter_next", "backgroud_loop", 
 ]
 
 from abc import ABC, abstractmethod
 from asyncio import create_task, sleep as async_sleep, to_thread
 from builtins import map as _map, filter as _filter, zip as _zip
-from collections import deque
+from collections import defaultdict, deque
 from collections.abc import (
-    AsyncIterable, AsyncIterator, Awaitable, Buffer, Callable, Coroutine, Generator, 
-    Iterable, Iterator, MutableSet, Sequence, 
+    AsyncIterable, AsyncIterator, Awaitable, Buffer, Callable, Collection, Container, 
+    Coroutine, Generator, Iterable, Iterator, Mapping, MutableMapping, MutableSet, 
+    MutableSequence, Sequence, ValuesView, 
 )
 from contextlib import asynccontextmanager, contextmanager, ExitStack, AsyncExitStack
+from copy import copy
 from dataclasses import dataclass
 from itertools import batched, pairwise
 from inspect import isawaitable, iscoroutinefunction
+from sys import _getframe
 from _thread import start_new_thread
 from time import sleep, time
-from typing import overload, Any, AsyncContextManager, ContextManager, Literal
+from types import FrameType
+from typing import cast, overload, Any, AsyncContextManager, ContextManager, Literal
 
 from asynctools import (
     async_filter, async_map, async_reduce, async_zip, async_batched, ensure_async, ensure_aiter, 
+    collect as async_collect, 
 )
 from decotools import optional
 from texttools import format_time
@@ -47,6 +53,12 @@ class YieldBase(ABC):
         ...
 
 
+class Reraised(BaseException):
+
+    def __init__(self, exc: BaseException, /):
+        self.exception = exc
+
+
 class Return(YieldBase):
     yield_type = 0
 
@@ -94,7 +106,7 @@ async def async_foreach(
     iterable: Iterable | AsyncIterable, 
     /, 
     *iterables: Iterable | AsyncIterable, 
-    threaded: bool = True, 
+    threaded: bool = False, 
 ):
     value = ensure_async(value, threaded=threaded)
     if iterables:
@@ -125,7 +137,7 @@ async def async_through(
     iterable: Iterable | AsyncIterable, 
     /, 
     take_while: None | Callable = None, 
-    threaded: bool = True, 
+    threaded: bool = False, 
 ):
     iterable = ensure_aiter(iterable, threaded=threaded)
     if take_while is None:
@@ -175,7 +187,7 @@ async def async_flatten(
     iterable: Iterable | AsyncIterable, 
     /, 
     exclude_types: type | tuple[type, ...] = (Buffer, str), 
-    threaded: bool = True, 
+    threaded: bool = False, 
 ) -> AsyncIterator:
     async for e in ensure_aiter(iterable, threaded=threaded):
         if isinstance(e, (Iterable, AsyncIterable)) and not isinstance(e, exclude_types):
@@ -185,6 +197,168 @@ async def async_flatten(
             yield e
 
 
+@overload
+def collect[K, V](
+    iterable: Iterable[tuple[K, V]] | Mapping[K, V], 
+    /, 
+    rettype: Callable[[Iterable[tuple[K, V]]], MutableMapping[K, V]], 
+) -> MutableMapping[K, V]:
+    ...
+@overload
+def collect[T](
+    iterable: Iterable[T], 
+    /, 
+    rettype: Callable[[Iterable[T]], Collection[T]] = list, 
+) -> Collection[T]:
+    ...
+@overload
+def collect[K, V](
+    iterable: AsyncIterable[tuple[K, V]], 
+    /, 
+    rettype: Callable[[Iterable[tuple[K, V]]], MutableMapping[K, V]], 
+) -> Coroutine[Any, Any, MutableMapping[K, V]]:
+    ...
+@overload
+def collect[T](
+    iterable: AsyncIterable[T], 
+    /, 
+    rettype: Callable[[Iterable[T]], Collection[T]] = list, 
+) -> Coroutine[Any, Any, Collection[T]]:
+    ...
+def collect(
+    iterable: Iterable | AsyncIterable | Mapping, 
+    /, 
+    rettype: Callable[[Iterable], Collection] = list, 
+) -> Collection | Coroutine[Any, Any, Collection]:
+    if not isinstance(iterable, Iterable):
+        return async_collect(iterable, rettype)
+    return rettype(iterable)
+
+
+@overload
+def group_collect[K, V, C: Container](
+    iterable: Iterable[tuple[K, V]], 
+    mapping: None = None, 
+    factory: None | C | Callable[[], C] = None, 
+) -> dict[K, C]:
+    ...
+@overload
+def group_collect[K, V, C: Container, M: MutableMapping](
+    iterable: Iterable[tuple[K, V]], 
+    mapping: M, 
+    factory: None | C | Callable[[], C] = None, 
+) -> M:
+    ...
+@overload
+def group_collect[K, V, C: Container](
+    iterable: AsyncIterable[tuple[K, V]], 
+    mapping: None = None, 
+    factory: None | C | Callable[[], C] = None, 
+) -> Coroutine[Any, Any, dict[K, C]]:
+    ...
+@overload
+def group_collect[K, V, C: Container, M: MutableMapping](
+    iterable: AsyncIterable[tuple[K, V]], 
+    mapping: M, 
+    factory: None | C | Callable[[], C] = None, 
+) -> Coroutine[Any, Any, M]:
+    ...
+def group_collect[K, V, C: Container, M: MutableMapping](
+    iterable: Iterable[tuple[K, V]] | AsyncIterable[tuple[K, V]], 
+    mapping: None | M = None, 
+    factory: None | C | Callable[[], C] = None, 
+) -> dict[K, C] | M | Coroutine[Any, Any, dict[K, C]] | Coroutine[Any, Any, M]:
+    if not isinstance(iterable, Iterable):
+        return async_group_collect(iterable, mapping, factory)
+    if factory is None:
+        if isinstance(mapping, defaultdict):
+            factory = mapping.default_factory
+        elif mapping:
+            factory = type(next(iter(ValuesView(mapping))))
+        else:
+            factory = cast(type[C], list)
+    elif callable(factory):
+        pass
+    elif isinstance(factory, Container):
+        factory = cast(Callable[[], C], lambda _obj=factory: copy(_obj))
+    else:
+        raise ValueError("can't determine factory")
+    factory = cast(Callable[[], C], factory)
+    if isinstance(factory, type):
+        factory_type = factory
+    else:
+        factory_type = type(factory())
+    if issubclass(factory_type, MutableSequence):
+        add = getattr(factory_type, "append")
+    else:
+        add = getattr(factory_type, "add")
+    if mapping is None:
+        mapping = cast(M, {})
+    for k, v in iterable:
+        try:
+            c = mapping[k]
+        except LookupError:
+            c = mapping[k] = factory()
+        add(c, v)
+    return mapping
+
+
+@overload
+async def async_group_collect[K, V, C: Container](
+    iterable: Iterable[tuple[K, V]] | AsyncIterable[tuple[K, V]], 
+    mapping: None = None, 
+    factory: None | C | Callable[[], C] = None, 
+    threaded: bool = False, 
+) -> dict[K, C]:
+    ...
+@overload
+async def async_group_collect[K, V, C: Container, M: MutableMapping](
+    iterable: Iterable[tuple[K, V]] | AsyncIterable[tuple[K, V]], 
+    mapping: M, 
+    factory: None | C | Callable[[], C] = None, 
+    threaded: bool = False, 
+) -> M:
+    ...
+async def async_group_collect[K, V, C: Container, M: MutableMapping](
+    iterable: Iterable[tuple[K, V]] | AsyncIterable[tuple[K, V]], 
+    mapping: None | M = None, 
+    factory: None | C | Callable[[], C] = None, 
+    threaded: bool = False, 
+) -> dict[K, C] | M:
+    iterable = ensure_aiter(iterable, threaded=threaded)
+    if factory is None:
+        if isinstance(mapping, defaultdict):
+            factory = mapping.default_factory
+        elif mapping:
+            factory = type(next(iter(ValuesView(mapping))))
+        else:
+            factory = cast(type[C], list)
+    elif callable(factory):
+        pass
+    elif isinstance(factory, Container):
+        factory = cast(Callable[[], C], lambda _obj=factory: copy(_obj))
+    else:
+        raise ValueError("can't determine factory")
+    factory = cast(Callable[[], C], factory)
+    if isinstance(factory, type):
+        factory_type = factory
+    else:
+        factory_type = type(factory())
+    if issubclass(factory_type, MutableSequence):
+        add = getattr(factory_type, "append")
+    else:
+        add = getattr(factory_type, "add")
+    if mapping is None:
+        mapping = cast(M, {})
+    async for k, v in iterable:
+        try:
+            c = mapping[k]
+        except LookupError:
+            c = mapping[k] = factory()
+        add(c, v)
+    return mapping
+
+
 def map(
     function: None | Callable, 
     iterable: Iterable | AsyncIterable, 
@@ -398,7 +572,7 @@ async def wrap_aiter[T](
     callnext: None | Callable[[T], Any] = None, 
     callenter: None | Callable[[Iterable[T] | AsyncIterable[T]], Any] = None, 
     callexit: None | Callable[[Iterable[T] | AsyncIterable[T], None | BaseException], Any] = None, 
-    threaded: bool = True, 
+    threaded: bool = False, 
 ) -> AsyncIterator[T]:
     callprev = ensure_async(callprev, threaded=threaded) if callable(callprev) else None
     callnext = ensure_async(callnext, threaded=threaded) if callable(callnext) else None
@@ -452,108 +626,180 @@ def cut_iter(
         yield stop, stop - start
 
 
-def run_gen_step[T](
-    gen_step: Generator[Any, Any, T] | Callable[[], Generator[Any, Any, T]], 
-    *, 
-    async_: bool = False, 
+def _get_async(back: int = 2) -> bool:
+    f: None | FrameType
+    f = _getframe(back)
+    f_globals = f.f_globals
+    f_locals  = f.f_locals
+    if f_locals is f_globals:
+        return f_locals.get("async_") or False
+    while f_locals is not f_globals:
+        if "async_" in f_locals:
+            return f_locals["async_"] or False
+        f = f.f_back
+        if f is None:
+            break
+        f_locals = f.f_locals
+    return False
+
+
+@overload
+def call_as_async[**Args, T: Coroutine](
+    func: Callable[Args, T], /, 
     threaded: bool = False, 
-    as_iter: bool = False, 
-) -> T:
+) -> Callable[Args, T]:
+    ...
+@overload
+def call_as_async[**Args, T](
+    func: Callable[Args, T], /, 
+    threaded: bool = False, 
+) -> Callable[Args, Coroutine[Any, Any, T]]:
+    ...
+def call_as_async[**Args, T](
+    func: Callable[Args, T], 
+    /, 
+    threaded: bool = False, 
+) -> Callable[Args, T] | Callable[Args, Coroutine[Any, Any, T]]:
+    if iscoroutinefunction(func):
+        return func
+    def wraps(*args, **kwds):
+        try:
+            return func(*args, **kwds)
+        except (StopIteration, StopAsyncIteration) as e:
+            raise Reraised(e) from e
+    if threaded:
+        async def wrapper(*args, **kwds):
+            return await to_thread(wraps, *args, **kwds)
+    else:
+        async def wrapper(*args, **kwds):
+            return wraps(*args, **kwds)
+    return wrapper
+
+
+def iter_gen_step(
+    gen_step: Generator | Callable[[], Generator], 
+):
     if callable(gen_step):
         gen = gen_step()
-        close = gen.close
     else:
         gen = gen_step
-        close = None
-    send = gen.send
+    send  = gen.send
     throw = gen.throw
+    close = gen.close
+    try:
+        value = send(None)
+        while True:
+            if isinstance(value, YieldBase):
+                raise StopIteration(value)
+            try:
+                if callable(value):
+                    value = value()
+            except BaseException as e:
+                value = throw(e)
+            else:
+                value = send(value)
+            yield value
+    except StopIteration as e:
+        value = e.value
+        try_call_me = True
+        if isinstance(value, YieldBase):
+            try_call_me = value.try_call_me
+            value = value.value
+        if callable(value) and try_call_me:
+            try:
+                value = value()
+            except BaseException as e:
+                try:
+                    value = throw(e)
+                except BaseException as e:
+                    raise Reraised(e) from e
+        yield value
+    except BaseException as e:
+        raise Reraised(e) from e
+    finally:
+        close()
+
+
+async def iter_gen_step_async(
+    gen_step: Generator | Callable[[], Generator], 
+    threaded: bool = False, 
+):
+    if callable(gen_step):
+        gen = gen_step()
+    else:
+        gen = gen_step
+    send:  Callable = call_as_async(gen.send, threaded=threaded)
+    throw: Callable = call_as_async(gen.throw, threaded=threaded)
+    close: Callable = call_as_async(gen.close, threaded=threaded)
+    try:
+        value = await send(None)
+        while True:
+            if isinstance(value, YieldBase):
+                raise StopIteration(value)
+            try:
+                if callable(value):
+                    value = await call_as_async(value, threaded=threaded)()
+                if isawaitable(value):
+                    value = await value
+            except BaseException as e:
+                if isinstance(e, Reraised):
+                    e = e.exception
+                value = await throw(e)
+            else:
+                value = await send(value)
+            yield value
+    except BaseException as e:
+        if isinstance(e, Reraised):
+            e = e.exception
+        if isinstance(e, StopIteration):
+            value = e.value
+            identity = False
+            try_call_me = True
+            if isinstance(value, YieldBase):
+                identity = value.identity
+                try_call_me = value.try_call_me
+                value = value.value
+            try:
+                if callable(value) and try_call_me:
+                    value = await call_as_async(value, threaded=threaded)()
+                if not identity and isawaitable(value):
+                    value = await value
+            except BaseException as e:
+                try:
+                    value = await throw(e)
+                except BaseException as e:
+                    raise Reraised(e) from e
+            yield value
+        else:
+            raise
+    finally:
+        await close()
+
+
+def run_gen_step[T](
+    gen_step: Generator | Callable[[], Generator], 
+    *, 
+    async_: None | bool = None, 
+    threaded: bool = False, 
+):
+    if async_ is None:
+        async_ = _get_async()
     if async_:
         async def process():
             try:
-                if threaded:
-                    value = await to_thread(send, None)
-                else:
-                    value = send(None)
-                while True:
-                    if isinstance(value, YieldBase):
-                        raise StopIteration(value)
-                    try:
-                        if callable(value):
-                            value = value()
-                        if isawaitable(value):
-                            value = await value
-                    except BaseException as e:
-                        if threaded:
-                            value = await to_thread(throw, e)
-                        else:
-                            value = throw(e)
-                    else:
-                        if threaded:
-                            value = await to_thread(send, value)
-                        else:
-                            value = send(value)
-            except StopIteration as e:
-                value = e.value
-                identity = False
-                try_call_me = True
-                if isinstance(value, YieldBase):
-                    identity = value.identity
-                    try_call_me = value.try_call_me
-                    value = value.value
-                if callable(value) and try_call_me:
-                    value = value()
-                if not identity and isawaitable(value):
-                    value = await value
+                async for value in iter_gen_step_async(gen_step, threaded=threaded):
+                    pass
                 return value
-            finally:
-                if close is not None:
-                    if threaded:
-                        await to_thread(close)
-                    else:
-                        close()
-        result = process()
-        if as_iter:
-            async def wrap(result):
-                iterable = await result
-                try:
-                    iterable = aiter(iterable)
-                except TypeError:
-                    for val in iter(iterable):
-                        if isawaitable(val):
-                            val = await val
-                        yield val
-                else:
-                    async for val in iterable:
-                        yield val
-            result = wrap(result)
-        return result
+            except Reraised as e:
+                raise e.exception
+        return process()
     else:
         try:
-            value = send(None)
-            while True:
-                if isinstance(value, YieldBase):
-                    raise StopIteration(value)
-                try:
-                    if callable(value):
-                        value = value()
-                except BaseException as e:
-                    value = throw(e)
-                else:
-                    value = send(value)
-        except StopIteration as e:
-            value = e.value
-            try_call_me = True
-            if isinstance(value, YieldBase):
-                try_call_me = value.try_call_me
-                value = value.value
-            if callable(value) and try_call_me:
-                value = value()
-            if as_iter:
-                value = iter(value)
+            for value in iter_gen_step(gen_step):
+                pass
             return value
-        finally:
-            if close is not None:
-                close()
+        except Reraised as e:
+            raise e.exception
 
 
 @overload
@@ -561,7 +807,15 @@ def run_gen_step_iter(
     gen_step: Generator | Callable[[], Generator], 
     threaded: bool = False, 
     *, 
-    async_: Literal[False] = False, 
+    async_: None = None, 
+) -> Iterator | AsyncIterator:
+    ...
+@overload
+def run_gen_step_iter(
+    gen_step: Generator | Callable[[], Generator], 
+    threaded: bool = False, 
+    *, 
+    async_: Literal[False], 
 ) -> Iterator:
     ...
 @overload
@@ -576,37 +830,38 @@ def run_gen_step_iter(
     gen_step: Generator | Callable[[], Generator], 
     threaded: bool = False, 
     *, 
-    async_: bool = False, 
+    async_: None | bool = None, 
 ) -> Iterator | AsyncIterator:
+    if async_ is None:
+        async_ = _get_async()
     if callable(gen_step):
         gen = gen_step()
-        close = gen.close
     else:
         gen = gen_step
-        close = None
-    send = gen.send
-    throw = gen.throw
+    send:  Callable = gen.send
+    throw: Callable = gen.throw
+    close: Callable = gen.close
     if async_:
+        send  = call_as_async(send, threaded=threaded)
+        throw = call_as_async(throw, threaded=threaded)
+        close = call_as_async(close, threaded=threaded)
         async def process():
             async def extract(value):
                 identity = False
                 try_call_me = True
                 yield_type = -1
                 if isinstance(value, YieldBase):
-                    identity = value.identity
+                    identity    = value.identity
                     try_call_me = value.try_call_me
-                    yield_type = value.yield_type
-                    value = value.value
+                    yield_type  = value.yield_type
+                    value       = value.value
                 if try_call_me and callable(value):
-                    value = value()
+                    value = await call_as_async(value, threaded=threaded)()
                 if not identity and isawaitable(value):
                     value = await value
                 return yield_type, value
             try:
-                if threaded:
-                    value = await to_thread(send, None)
-                else:
-                    value = send(None)
+                value = await send(None)
                 while True:
                     try:
                         yield_type, value = await extract(value)
@@ -617,29 +872,31 @@ def run_gen_step_iter(
                                 async for val in ensure_aiter(value, threaded=threaded):
                                     yield val
                     except BaseException as e:
-                        if threaded:
-                            value = await to_thread(throw, e)
-                        else:
-                            value = throw(e)
+                        if isinstance(e, Reraised):
+                            e = e.exception
+                        value = await throw(e)
                     else:
-                        if threaded:
-                            value = await to_thread(send, value)
-                        else:
-                            value = send(value)
-            except StopIteration as e:
-                yield_type, value = await extract(e.value)
-                match yield_type:
-                    case 1:
-                        yield value
-                    case 2:
-                        async for val in ensure_aiter(value, threaded=threaded):
-                            yield val
+                        value = await send(value)
+            except BaseException as e:
+                if isinstance(e, Reraised):
+                    e = e.exception
+                if isinstance(e, StopIteration):
+                    try:
+                        yield_type, value = await extract(e.value)
+                        match yield_type:
+                            case 1:
+                                yield value
+                            case 2:
+                                async for val in ensure_aiter(value, threaded=threaded):
+                                    yield val
+                    except BaseException as e:
+                        if isinstance(e, Reraised):
+                            e = e.exception
+                        await throw(e)
+                else:
+                    raise
             finally:
-                if close is not None:
-                    if threaded:
-                        await to_thread(close)
-                    else:
-                        close()
+                await close()
     else:
         def process():
             def extract(value, /):
@@ -669,17 +926,19 @@ def run_gen_step_iter(
                     else:
                         value = send(value)
             except StopIteration as e:
-                yield_type, value = extract(e.value)
-                match yield_type:
-                    case 1:
-                        yield value
-                    case 2:
-                        yield from value
-                    case _:
-                        return value
+                try:
+                    yield_type, value = extract(e.value)
+                    match yield_type:
+                        case 1:
+                            yield value
+                        case 2:
+                            yield from value
+                        case _:
+                            return value
+                except BaseException as e:
+                    throw(e)
             finally:
-                if close is not None:
-                    close()
+                close()
     return process()
 
 
@@ -687,9 +946,11 @@ def run_gen_step_iter(
 def as_gen_step(
     func: Callable, 
     /, 
-    async_: bool = False, 
+    async_: None | bool = None, 
     iter: bool = False, 
 ) -> Callable:
+    if async_ is None:
+        async_ = _get_async()
     def wrapper(*args, **kwds):
         if iter:
             return run_gen_step_iter(func(*args, **kwds), async_=async_) # type: ignore
